/*    Copyright (C) 2005 XIAO, Gang of Universite de Nice - Sophia Antipolis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

	/* Userspace interface for Linux sysmask */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/net.h>
#include <linux/capability.h>
#include "syscalls.h"
#include "sysmask.h"
#include "smk.h"

static inline _syscall3(long,sysmask,long,x,void *,y,long,z)

int isquery;
struct {
    struct sysmask_struct smk;
    struct smk_session ses[SMK_SESSIONS];
} query_data;
#define psmk query_data.smk
int pquery;

#include "masktab.c"
#include "smklib.c"

char usage[]="\
Usage: sysmask [options] masks [command [arguments...]]\n\
Options:\n\
  -q		Query the mask setup.\n\
  -qa		Detailed query with list of unmasked syscalls.\n\
  -t trigger	Triggered mask; trigger may be a name or value.\n\
  -n tcount	Used with trigger, set trigger count to tcount.\n\
  -h		Print this message.\n\
  -v		Print sysmask version.\n\
\n\
masks must be a non-empty list of mask names, token names, or \
process ids.\n\
\n\
This program may be called in chain.\n\n\
";

void putversion(void)
{
    printf("Linux sysmask %s, kernel protocol version %d.\n",
	   VERSION, SYSMASK_VERSION);
}

void readfile(char *fname,char *buf,long bufsize)
{
    struct stat st;
    int f;
    long l,l2;

    buf[0]=0;
    if(stat(fname,&st)) return;
    l=st.st_size;
    if(l<0 || l>=bufsize) {
	fprintf(stderr,"%s: file too long.\n",fname);
	return;
    }
    f=open(fname,0); if(f<0) {
	fprintf(stderr,"%s: cannot open. %s\n",fname,strerror(errno));
	return;
    }
    l2=read(f,buf,l); if(l2!=l) {
	fprintf(stderr,"%s: read() error. %s\n",fname,strerror(errno));
	close(f); l=0;
    }
    buf[l]=0; close(f);
    _tolinux(buf);    
}

int checkdaemon(void)
{
    struct sysmask_struct smk;
    struct stat st;
    int i;
    getsysmask(&smk);	/* token mask? */
    if(smk.active.m1 || smk.active.m2 || smk.active.cap || smk.token) return 0;
    if(stat("/sbin/smkd",&st)!=0) return 0;	/* no daemon script */
    i=fork();
    if(i==0) {	/* child */
	execl("/sbin/smkd","/sbin/smkd",NULL);
	exit(127);
    }
    else if(i>0) wait(&i);
    return 1;
}

#include "tokens.c"

void query_process(char *p)
{
    int i;
    if(pquery) {
	fprintf(stderr,"sysmask warning: double process query %s ignored.\n",p);
	return;
    }
    pquery=atoi(p);
    if(pquery<=0) return;
    psmk.token=pquery; i=sysmask(SMK_GETMASK2,&query_data,sizeof(query_data));
    if(i<0) {
	fprintf(stderr,"sysmask warning: failed to get mask of process %s. %s\n",p,
		strerror(-i));
	pquery=0; return;
    }
}

int query(char querychar,struct sysmask_struct *mst)
{
    struct sysmask_struct smk;
    int t,i,j,k;

    if(pquery) {
	memmove(&smk,mst,sizeof(smk));
	printf("Taking mask from process %d.\n",pquery);
    }
    else {
	i=getsysmask(&smk);
	if(i<0) return -1;
    }
    printf("Token: ");
    i=0;
    t=smk.token&0xFFFF; if(t) {
	i++;
	if(tokenn[t]) printf("%s ",tokenn[t]);
	else printf("s%d ",t);
    }
    t=(smk.token>>16)&0xFFFF; if(t) {
	i++;
	if(tokenn[(1<<16)+t]) printf("%s ",tokenn[(1<<16)+t]);
	else printf("u%d ",t);
    }
    if(!i) printf("none. ");
    j=0;
    if(query_data.ses[0].id) {
	j++;
	printf("\nSystem session %lu, %lu, %lu, %lu. ",query_data.ses[0].id,
	       query_data.ses[0].killid,
	       (unsigned long) query_data.ses[0].rstatus[SMK_RLIM_NPROC].counter,
	       (unsigned long) query_data.ses[0].rstatus[SMK_RLIM_FORK].counter);
    }
    if(query_data.ses[1].id) {
	if(!j) printf("\n");
	printf("User session %lu, %lu, %lu, %lu. ",query_data.ses[1].id,
	       query_data.ses[1].killid,
	       (unsigned long) query_data.ses[1].rstatus[SMK_RLIM_NPROC].counter,
	       (unsigned long) query_data.ses[1].rstatus[SMK_RLIM_FORK].counter);
    }
    printf("\nsm1=%lX sm2=%lX  cap=%lX\n",
	   smk.active.m1,smk.active.m2,smk.active.cap);
    for(i=0;i<SMK_EXCEPTS && smk.except[i]; i++);
    if(i>0) {
	printf("Excepted syscalls: ");
	for(j=0;j<i;j++) {
	    if(smk.except[j]==0) continue;
	    if(smk.except[j]==SMKD_EXCEPT) {
		printf("daemon "); continue;
	    }
	    if(smk.except[j]<=NR_sysmasks && syscall_names[smk.except[j]-1].name)
	      printf("%s ",syscall_names[smk.except[j]-1].name);
	    else printf("%d ",smk.except[j]-1);
	}
	printf("\n");
    }
    if(smk.trigger && smk.tcnt>0) {
	for(i=0; i<trcnt && triggers[i].val!=smk.trigger; i++);
	if(i>=trcnt) printf("Unknown trigger %d\n",smk.trigger);
	else {
	    t=smk.tcnt; if(t<1) t=1;
	    printf("Trigger %s, count %d.\n",
		   triggers[i].name, t);
	    smk.active.m1|=smk.triggered.m1;
	    smk.active.m2|=smk.triggered.m2;
	    smk.active.cap|=smk.triggered.cap;
	    printf("tm1=%lX tm2=%lX tcap=%lX\n",
		   smk.active.m1,smk.active.m2,smk.active.cap);
	}
    }
    else printf("No trigger.\n");
    if(querychar=='a') 
      printf("Non-masked syscalls: (stars indicate risk level)\n");
    for(i=j=t=0;i<NR_sysmasks;i++) {
	if(smk.active.m1 & sysmask_tab[i]) {
	    int n;
	    i++;
	    for(n=0;n<SMK_EXCEPTS && smk.except[n]!=i;n++);
	    i--;
	    if(n>=SMK_EXCEPTS) continue;
	}
		/* non-implemented */
	if(i==17 || i==31 || i==32 || i==35 || i==44 ||
	   i==53 || i==58 || i==98 || i==112 || i==137 ||
	   i==188 || i==189 || i==222 || i==223 || i==251) continue;
	if(i==__NR_execve && (smk.active.m2&SM2_EXECVE)) continue;
	if(i==__NR_creat && (smk.active.m2&SM2_CREAT)) continue;
	if(i==__NR_modify_ldt && (smk.active.m2&SM2_LDT)) continue;
	if((smk.active.m2&SM2_SUID) &&
	   (i==23 || i==46 || i==70 || i==71 || i==81 || i==164 || i==170 ||
	    i==203 || i==204 || i==206 || i==208 || i==210 || i==213 || i==214))
	  continue;
	if(syscall_names[i].name) {
	    k=strlen(syscall_names[i].stars);
	    if(t==k) j++;
	    if(t<k) {t=k; j=1;}
	    if(querychar=='a')
	      printf("%s	%s\n", syscall_names[i].stars,
		     syscall_names[i].name);
	}
    }
    printf("\nMaximal risk level: %d (%d calls).\n", t,j);
    return 0;
}

int main(int argc,char *argv[])
{
    char *args[1024];
    int i,k,tcnt,trigger,reverse;
    struct sysmask_maskset mask;
    struct sysmask_struct mst;
    char maskstr[1024];
    long token;
    char *p1, *p2, querychar=0;

    if(argc<2) goto help;
    trigger=tcnt=reverse=isquery=pquery=0;
    memset(&psmk,0,sizeof(psmk));
    get_syscall_names();
    tcnt=1;
    for(k=1; k<argc; k++) {
	if(argv[k][0]!='-' || argv[k][1]==0) break;
	p1=argv[k]+1; p2=p1+1;
	if(*p2==0 && (*p1 == 't' || *p1 == 'n')) {
	    if(k>=argc-1) {
		fprintf(stderr,"%s: value required for option -%c.\n",
			argv[0], *p1);
		exit(1);
	    }
	    p2=argv[++k];
	}
	switch(*p1) {
	    case '-': {
		if(strcmp(p1+1,"help")!=0) goto unknown_option;
	    }
	    case 'h': {
		help:
		putversion();
		printf("%s", usage); return 0;
	    }
	    case 't': {		/* trigger */
		trigger=gettrigger(p2);
		if(trigger==0) 
		  fprintf(stderr,"%s: warning: unknown trigger %s.\n",
			  argv[0],p2);
		break;
	    }
	    case 'n': {		/* trigger count */
		i=atoi(p2); if(i<=0 || i>30000) {
		    fprintf(stderr,"%s: invalid trigger count %s.\n",
			    argv[0],p2);
		    exit(1);
		}
		tcnt=i; break;
	    }
	    case 'q': {
		isquery=1; querychar=*p2; break;
	    }
	    case 'v': {putversion(); return 0;}
	    case 'z': {
		printf("%ld\n",sysmask(SMK_SESSIONCNT,NULL,0));
		return 0;
	    }
	    default: {
		unknown_option:
		fprintf(stderr,"%s: unknown option -%c.\n",argv[0],*p1);
		exit(1);
	    }
	}
    }
    if(k<argc) snprintf(maskstr,sizeof(maskstr),"%s",argv[k++]);
    else {
	maskstr[0]=0;
	if(!isquery) 
	  fprintf(stderr,"%s: warning: mask string not found. Assumes \"none\".\n",
		  argv[0]);
    }
    i=sysmask(SMK_VERSION,NULL,0);
    if(i!=SYSMASK_VERSION) {
	if(i<0 && errno == ENOSYS) {
	    fprintf(stderr,"%s: kernel has no sysmask support. Protection ignored.\n",argv[0]);
	    goto exec;
	}
	fprintf(stderr,"%s warning: wrong kernel sysmask version %d ignored.\n",argv[0],i);
	if(i==-1) fprintf(stderr,"%s\n",strerror(errno));
	goto exec;
    }
    memset(&mst,0,sizeof(mst)); memset(&mask,0,sizeof(mask));
    token=0; tokentime=0; gettokens();
    for(p1=maskstr; *p1; p1=p2) {
	while(*p1 && !isalnum(*p1)) p1++;
	for(p2=p1; isalnum(*p2) || *p2=='_'; p2++) *p2=tolower(*p2);
	if(*p2) *p2++=0;
	if(*p1==0) continue;
	if(trigger==0) {
	    if(*p1=='s' || *p1=='u') {
		char *pt;
		for(pt=p1+1; myisdigit(*pt); pt++);
		if(*pt==0) {	/* numeric definition of a token */
		    i=atoi(p1+1);
		    if(i>0 && i<0x10000) {
			if(*p1=='u') i<<=16;
			addtoken(&token,i);
		    }
		    continue;
		}
	    }
	    i=search_list(tokens,tokencnt,sizeof(tokens[0]),p1);
	    if(i>=0) {
		addtoken(&token,token_expand(tokens[i].n));
		continue;
	    }
	}
	if(*p1=='p' && myisdigit(p1[1])) {
	    query_process(p1+1);
	    continue;
	}
	onemask(p1,&mask,mst.except);
    }
    if((mask.m1|mask.m2|mask.cap|token)!=0 || pquery) {
	if(trigger) {
	    memmove(&mst.triggered,&mask,sizeof(mask));
	    mst.trigger=trigger;
	    mst.tcnt=tcnt;
	}
	else memmove(&mst.active,&mask,sizeof(mask));
	if(pquery) {
	    addmask(&mst,&psmk);
	    addtoken(&token,psmk.token&0xFFFF);
	    addtoken(&token,psmk.token&0xFFFF0000);
	}
	mst.token=token;
	i=setsysmask(&mst);
	if(i) {
	    fprintf(stderr,"%s: sysmask() failed: %s\n",argv[0],strerror(errno));
	    exit(1);
	}
    }
    else if(!isquery) {
	fprintf(stderr,"%s: warning: no valid mask defined.\n",argv[0]);
    }
    if(isquery) {
	if(query(querychar,&mst)<0) {
	    fprintf(stderr,"%s: query failed. %s\n",
		    argv[0], strerror(errno));
	    return 1;
	}
	else return 0;
    }
    if(mst.token==1 && geteuid()==0 && getuid()==0 &&
       !mst.active.m1 && !mst.active.m2 && !mst.active.cap) {	/* init */
	if(checkdaemon()) setsysmask(&mst);
    }
    exec:
    if(k<argc) {
	for(i=0;i<1000 && i<argc-k; i++) args[i]=argv[i+k];
	args[i]=NULL;
    }
    else {args[0]="/bin/sh"; args[1]=NULL;}
    if(strchr(args[0],'/')) execv(args[0],args); else execvp(args[0],args);
    fprintf(stderr,"%s: cannot execute %s. %s\n", argv[0], args[0],strerror(errno));
    return 127;
}

